(function (root, factory) {
    // requirejs?
    if (typeof define === 'function') {
        define('Basthon', ['BasthonBase'], factory);
    // TypeScript?
    } else if (typeof exports === 'object') {
        module.exports = factory();
    // default goes to global namespace
    } else {
        root.Basthon = factory(BasthonBase);
    }
}(this, function (BasthonBase) {
    'use strict';

    class Basthon extends BasthonBase {
        constructor() {
            super();
            this.execution_count = 0;
            this._sqlJsUrl = "https://cdnjs.cloudflare.com/ajax/libs/sql.js/1.5.0/sql-wasm.min.js";
        };

        language () { return "sql"; };
        languageName () { return "SQL"; };
        moduleExts () { return ['sql', 'db', 'sqlite']; };

        async launch () {
            // avoid conflict with requirejs and use it when available.
            try {
                if( typeof requirejs !== 'undefined' ) {
                    requirejs.config({paths: {'sql-wasm': this._sqlJsUrl.slice(0, -3)}});
                    await new Promise(function (resolve, reject) {
                        require(['sql-wasm'], resolve, reject);
                    });
                } else {
                    await this.loadScript(this._sqlJsUrl);
                }
            } catch (error) {
                console.log(error);
                console.error("Can't load sql.js");
            }

            const SQL = await window.initSqlJs({
                locateFile: file => `https://sql.js.org/dist/${file}`
            });

            this.db = new SQL.Database();

            await this.pageLoad();
        };

        troncate_table (json) {
            const max_size = 150;
            if( json.values.length >= max_size ) {
                const half_length = Math.floor(max_size / 3);
                const head = json.values.slice(0, half_length);
                const tail = json.values.slice(-half_length);
                json.values = head.concat([json.columns.map(() => '⋮ ')], tail);
            }
        };

        html_table (json) {
            this.troncate_table(json);
            const columns = json.columns.map((c) => `<th>${c}</th>`).join('');
            const head = `<thead><tr style="text-align: right;">${columns}</tr></thead>`;
            const values = json.values.map((r) => '<tr>' + r.map((v) => `<td>${v}</td>`).join('') + '</tr>').join('');
            const body = `<tbody>${values}</tbody>`;
            return `<table class="dataframe" border="1">${head}${body}</table>`;
        };

        text_table (json) {
            this.troncate_table(json);
            const head = json.columns.join('\t');
            const values = json.values.map((r) => r.join('\t')).join('\n');
            return `${head}\n${values}`;
        };
        
        async evalAsync (code, outCallback, errCallback, data=null) {
            // force interactivity in all modes
            data.interactive = true;

            this.execution_count++;

            let result = this.db.exec(code);

            // return result
            if( typeof result !== 'undefined' ) {
                if( result.length == 0 )
                    result = undefined
                else
                    result = {'text/plain': this.text_table(result[0]),
                              'text/html': this.html_table(result[0])};
            }
            return [result, this.execution_count];
        };

        ps1 () { return "sql> "; };

        ps2 () { return "...> "; };

        restart () {
            // warning: we should probably reload module files...
            if( this.db ) this.db.close();
            this.execution_count = 0;
            this.db = new SQL.Database();
        };

        more (source) { return false; };
          
        complete (code) { return []; };

        putFile (filename, content) {
            console.error(`Fichier ${filename} not added since putFile has no mean in the SQL context.`);
        };

        putModule (filename, content) {
            const ext = filename.split('/').pop();
            switch( ext ) {
            case 'sql':
                this.db.run(content);
                break;
            case 'sqlite':
            case 'db':
                if( this.db ) this.db.close();
                this.db = new SQL.Database(content);
                break;
            default:
                throw {message: "Only '.sql', '.db' and '.sqlite' files supported."};
            }
        };
    }

    return new Basthon();
}));
